/***
 * Cecilia ADL Compiler
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.adl;

import java.io.File;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.Factory;
import org.objectweb.fractal.adl.FactoryFactory;
import org.objectweb.fractal.adl.util.Printer;
import org.objectweb.fractal.cecilia.adl.implementations.LanguageLoader;
import org.objectweb.fractal.cecilia.adl.plugin.AbstractPluginManager;

/**
 * A class to compile a component. Usage: Launcher &lt;definition&gt; where
 * &lt;definition&gt; is the name of the component to be compiled.
 */
public class Launcher extends AbstractLauncher {

  /**
   * The name of the {@link System#getProperty(String) system property} that
   * specify the backend to use for the plugin factory.
   */
  public static final String PLUGIN_FACTORY_BACKEND_PROPERTY_NAME = AbstractPluginManager.PLUGIN_FACTORY_BACKEND;

  /**
   * The default value of the {@link #PLUGIN_FACTORY_BACKEND_PROPERTY_NAME}
   * property.
   */
  public static final String DEFAULT_PLUGIN_FACTORY_BACKEND       = "Java";

  /** The name of the default ADL of the compiler. */
  public static final String CECILIA_BASIC_COMPILER_ADL           = "org.objectweb.fractal.cecilia.adl.CeciliaBasicFactory";

  private final CmdOption    compilerADLOpt                       = new CmdOption(
                                                                      new String[]{"compiler-adl"},
                                                                      "Specify the name of the ADL of the compiler itself",
                                                                      CECILIA_BASIC_COMPILER_ADL);

  private final CmdOption    compilerCmdOpt                       = new CmdOption(
                                                                      new String[]{
      "compiler-command", "cc"                                        },
                                                                      "the command of the C compiler (default is 'gcc')",
                                                                      "gcc");

  private final CmdOption    defsOpt                              = new CmdAppendOption(
                                                                      new String[]{
      "defs", "D"                                                     },
                                                                      "the -D compiler directives");
  private final CmdOption    cFlagsOpt                            = new CmdAppendOption(
                                                                      new String[]{
      "c-flags", "cf"                                                 },
                                                                      "the c-flags compiler directives");

  private final CmdOption    includePathOpt                       = new CmdPathOption(
                                                                      new String[]{
      "inc-path", "I"                                                 },
                                                                      "the -I compiler directives");

  private final CmdOption    linkerCmdOpt                         = new CmdOption(
                                                                      new String[]{
      "linker-command", "ld"                                          },
                                                                      "the command of the linker (default is 'gcc')",
                                                                      "gcc");

  private final CmdOption    executableFormat                     = new CmdOption(
                                                                      new String[]{"executable-format"},
                                                                      "the binary format associated to \"executable\" output format (default is 'elf')",
                                                                      "elf");

  private final CmdOption    ldFlagsOpt                           = new CmdAppendOption(
                                                                      new String[]{
      "ld-flags", "lf"                                                },
                                                                      "the ld-flags compiler directives");
  private final CmdOption    ldPathOpt                            = new CmdPathOption(
                                                                      new String[]{
      "ld-path", "-L"                                                 },
                                                                      "the -L compiler directives");

  private final CmdOption    archiveCmdOpt                        = new CmdOption(
                                                                      new String[]{
      "archive-command", "ar"                                         },
                                                                      "the command of the archive tool, (default is 'ar')",
                                                                      "ar");

  private final CmdOption    linkerScriptOpt                      = new CmdOption(
                                                                      new String[]{
      "linker-script", "-T"                                           },
                                                                      "linker script");

  private final CmdOption    executableNameCmdOpt                 = new CmdOption(
                                                                      new String[]{
      "executable-name", "exec-name"                                  },
                                                                      "the name of the target executable file",
                                                                      "");

  Map<?, ?>                  compilerComponent;
  Map<String, Object>        compilerContext                      = new HashMap<String, Object>();
  Factory                    compiler;

  /**
   * Creates a new Compiler launcher context.
   * 
   * @param outDir the output directory
   * @param sources the source path.
   */
  public Launcher(final String outDir, final String sources) {
    this(new String[]{"-o=" + outDir, "-src-path=" + sources,
        "this_is_not_an_adl"}, false);
  }

  /**
   * Creates a new Compiler launcher context.
   * 
   * @param options the options
   */
  public Launcher(final Map<String, String> options) {
    this(toArgs(options), false);
  }

  private static String[] toArgs(final Map<String, String> options) {
    final String[] params = new String[options.size()];
    int i = 0;
    for (final Map.Entry<String, String> option : options.entrySet()) {
      params[i] = option.getKey() + '=' + option.getValue();
      i++;
    }
    return params;
  }

  /**
   * Create the o.o.f.a.Factory class that will be used to read the application
   * definition. Subclasses of the Launcher may extend this.
   * 
   * @param compilerDefinition
   * @param compilerContext
   * @return
   * @throws ADLException
   * @throws Exception
   * @throws Exception
   */
  protected Factory createCompiler(final Map<String, Object> compilerContext) {
    /* instantiate the bootstrap factory with the Java backend */
    final Factory factory = FactoryFactory.getFactory();

    try {
      final String compilerAdlDefinition = compilerADLOpt.getValue();
      /* try to instantiate the CeciliaADL BasicFactory */
      compilerComponent = (Map<?, ?>) factory.newComponent(
          compilerAdlDefinition, compilerContext);
    } catch (final ADLException e) {
      internalError("Unable to instantiate Cecilia compiler", e, 101);
    }

    final Factory result = (Factory) compilerComponent.get("factory");

    return result;
  }

  protected Launcher(final String[] args, final boolean exitOnFailure) {
    super(exitOnFailure);

    if (Printer.isDebugEnable()) {
      for (final String arg : args) {
        Printer.debug("[arg] " + arg);
      }
    }

    setSpecificOptions(new CmdOption[]{compilerADLOpt, compilerCmdOpt, defsOpt,
        cFlagsOpt, includePathOpt, linkerCmdOpt, ldFlagsOpt, ldPathOpt,
        executableFormat, archiveCmdOpt, linkerScriptOpt, executableNameCmdOpt});

    parseArgs(args);

    // add to the compiler context default LanguageLoader arguments values
    compilerContext.put(LanguageLoader.DEFAULT_PPL_PROPERTY_NAME, System
        .getProperty(LanguageLoader.DEFAULT_PPL_PROPERTY_NAME,
            LanguageLoader.DEFAULT_PPL));

    compilerContext.put(LanguageLoader.DEFAULT_CPL_PROPERTY_NAME, System
        .getProperty(LanguageLoader.DEFAULT_CPL_PROPERTY_NAME,
            LanguageLoader.DEFAULT_CPL));

    compilerContext.put(LanguageLoader.DEFAULT_CPLS_PROPERTY_NAME, System
        .getProperty(LanguageLoader.DEFAULT_CPLS_PROPERTY_NAME,
            LanguageLoader.DEFAULT_CPLS));

    compilerContext.put(PLUGIN_FACTORY_BACKEND_PROPERTY_NAME, System
        .getProperty(PLUGIN_FACTORY_BACKEND_PROPERTY_NAME,
            DEFAULT_PLUGIN_FACTORY_BACKEND));

    /*
     * get the CeciliaADL Compiler (default implementation a pure Java
     * component)
     */

    this.compiler = createCompiler(compilerContext);

    final List<String> incPaths = new ArrayList<String>(sourcePaths);
    if (this.includePathOpt.getValue() != null)
      incPaths.addAll(parsePathList(includePathOpt.getValue()));
    StringBuffer sb = new StringBuffer();
    for (final String inc : incPaths) {
      sb.append(" -I").append(inc);
    }
    final String includePath = sb.toString();

    final List<String> ldPaths = new ArrayList<String>();
    if (ldPathOpt.getValue() != null)
      ldPaths.addAll(parsePathList(ldPathOpt.getValue()));
    sb = new StringBuffer();
    for (final String ld : ldPaths) {
      sb.append(" -L").append(ld);
    }
    final String ldPath = sb.toString();

    // add compiler arguments to context
    compilerContext.put(compilerCmdOpt.options[0], compilerCmdOpt.getValue());
    compilerContext.put(cFlagsOpt.options[0], cFlagsOpt.getValue());
    compilerContext.put(includePathOpt.options[0], includePath);
    compilerContext.put(defsOpt.options[0], defsOpt.getValue());

    compilerContext.put(linkerCmdOpt.options[0], linkerCmdOpt.getValue());
    compilerContext.put(ldFlagsOpt.options[0], ldFlagsOpt.getValue());
    compilerContext.put(ldPathOpt.options[0], ldPath);
    compilerContext.put(archiveCmdOpt.options[0], archiveCmdOpt.getValue());
    compilerContext.put(executableFormat.options[0], executableFormat
        .getValue());
    compilerContext.put(executableNameCmdOpt.options[0], executableNameCmdOpt
        .getValue());

    // add build directories to context
    compilerContext.put("adlBuildDirectory", adlBuildDir);
    compilerContext.put("idlBuildDirectory", idlBuildDir);
    compilerContext.put("objBuildDirectory", objBuildDir);

    // add source class loader in context
    compilerContext.put("classloader", sourceClassLoader);

    // add linker script to the context
    if (linkerScriptOpt.getValue() != null)
      compilerContext.put(linkerScriptOpt.options[0], linkerScriptOpt
          .getValue());

  }

  /**
   * Compile the given ADL.
   * 
   * @param adl the name of the ADL to compile.
   * @return the compiled file.
   * @throws ADLException if the compilation fails.
   */
  public File compile(final String adl) throws ADLException {
    adlName = adl;
    return compile();
  }

  private File compile() throws ADLException {
    final Map<Object, Object> context = new HashMap<Object, Object>(
        compilerContext);

    final Object result = this.compiler.newComponent(adlName, context);
    if (result instanceof Collection)
      return (File) ((Collection<?>) result).iterator().next();
    else
      return null;
  }

  @Override
  protected void printUsage(final PrintStream ps) {
    ps.println("Usage: Launcher [OPTIONS] <definition>");
    ps.println("  where <definition> is the name of the component to"
        + " be compiled,");
  }

  /**
   * Entry point.
   * 
   * @param args
   */
  public static void main(final String[] args) {
    doMain(new Launcher(args, true));
  }

  protected static void doMain(final Launcher launcher) {
    try {
      launcher.compile();
    } catch (final ADLException e) {
      e.printStackTrace();
// if (Printer.isDebugEnable()) {
// e.printStackTrace();
// } else {
// if (e instanceof TaskExecutionException) {
// final TaskExecutionException tee = (TaskExecutionException) e;
// final Throwable cause = tee.getCause();
// if (cause instanceof CompilationException) System.exit(-1);
// }

      final StringBuffer sb = new StringBuffer();
      sb.append(e.getMessage()).append('\n');
      Throwable cause = e.getCause();
      while (cause != null) {
        sb.append("caused by : ");
        if (cause instanceof ClassNotFoundException) {
          sb.append("Can't find IDL, ADL or source file: '").append(
              cause.getMessage()).append("'");
        } else if (cause instanceof ADLException) {
          sb.append(cause.getMessage());
        } else {
          final StringWriter sw = new StringWriter();
          cause.printStackTrace(new PrintWriter(sw));
          sb.append(sw.getBuffer());
        }
        sb.append('\n');
        cause = cause.getCause();
      }
      System.err.println(sb);
      System.exit(-1);
    }

  }

}
