/***
 * Fractal ADL Parser
 * Copyright (C) 2002-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.cecilia.adl.components;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.Compiler;
import org.objectweb.fractal.adl.Definition;
import org.objectweb.fractal.adl.components.Component;
import org.objectweb.fractal.adl.components.ComponentContainer;
import org.objectweb.fractal.adl.components.PrimitiveCompiler;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.task.core.TaskMap;

/**
 * Basic implementation of the {@link Compiler} interface. This implementation
 * delegates definition compilation requests to a {@link PrimitiveCompiler}.
 */
// TODO is this component class still useful ?
// The o.o.f.a.components.ComponentCompiler delegates to a set of
// primitiveCompiler. This one delegate to only one. we may try to merge these
// two class
public class ComponentCompiler implements Compiler, BindingController {

  // --------------------------------------------------------------------------
  // Client interface
  // --------------------------------------------------------------------------

  /** The name of the {@link PrimitiveCompiler} client interface. */
  public static final String PRIMITIVE_COMPILER_ITF = "primitive-compiler";

  /** The primitive compiler used by this compiler. */
  public PrimitiveCompiler   primitiveCompilerItf;

  // --------------------------------------------------------------------------
  // Implementation of the Compiler interface
  // --------------------------------------------------------------------------

  public void compile(final Definition definition, final TaskMap tasks,
      final Map<Object, Object> context) throws ADLException {
    if (definition instanceof ComponentContainer) {
      compile(new ArrayList<ComponentContainer>(),
          (ComponentContainer) definition, tasks, context);
    }
  }

  // --------------------------------------------------------------------------
  // Compilation methods
  // --------------------------------------------------------------------------

  void compile(final List<ComponentContainer> path,
      final ComponentContainer container, final TaskMap tasks,
      final Map<Object, Object> context) throws ADLException {
    path.add(container);
    for (final Component comp : container.getComponents()) {
      compile(path, comp, tasks, context);
    }
    path.remove(path.size() - 1);

    primitiveCompilerItf.compile(path, container, tasks, context);
  }

  // --------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // --------------------------------------------------------------------------

  public void bindFc(final String itfName, final Object o)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException {

    if (itfName == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (itfName.equals(PRIMITIVE_COMPILER_ITF)) {
      this.primitiveCompilerItf = (PrimitiveCompiler) o;
    } else {
      throw new NoSuchInterfaceException("There is no interface named '"
          + itfName + "'");
    }
  }

  public String[] listFc() {
    return new String[]{PRIMITIVE_COMPILER_ITF};
  }

  public Object lookupFc(final String itfName) throws NoSuchInterfaceException {
    if (itfName == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (itfName.equals(PRIMITIVE_COMPILER_ITF)) {
      return this.primitiveCompilerItf;
    } else {
      throw new NoSuchInterfaceException("There is no interface named '"
          + itfName + "'");
    }
  }

  public void unbindFc(final String itfName) throws NoSuchInterfaceException,
      IllegalBindingException, IllegalLifeCycleException {

    if (itfName == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (itfName.equals(PRIMITIVE_COMPILER_ITF)) {
      this.primitiveCompilerItf = null;
    } else {
      throw new NoSuchInterfaceException("There is no interface named '"
          + itfName + "'");
    }
  }
}
