/***
 * Cecilia ADL Compiler
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author: Matthieu Leclercq
 * Contributor: Alessio Pace
 */

package org.objectweb.fractal.cecilia.adl.components;

import java.util.List;
import java.util.Map;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.components.Component;
import org.objectweb.fractal.adl.components.ComponentContainer;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.task.codegen.api.FactoryProviderTask;
import org.objectweb.fractal.task.codegen.api.InstanceProviderTask;
import org.objectweb.fractal.task.codegen.api.RequireSourceCodeProviderTask;
import org.objectweb.fractal.task.codegen.lib.AbstractRecursiveSourceCodeProviderTask;
import org.objectweb.fractal.task.core.TaskMap;

/**
 * Adds a task to add sub component to composite component.
 */
public class ContentCompiler extends AbstractPrimitiveCompiler
    implements
      BindingController {

  // --------------------------------------------------------------------------
  // Client interface.
  // --------------------------------------------------------------------------

  /**
   * The name of the {@link ContentBuilder} client interface used by this
   * compiler.
   */
  public static final String BUILDER_ITF = "builder";

  /** The {@link ContentBuilder} used by this compiler. */
  protected ContentBuilder   builderItf;

  // --------------------------------------------------------------------------
  // Implementation of the PrimitiveCompiler interface
  // --------------------------------------------------------------------------

  public void compile(final List<ComponentContainer> path,
      final ComponentContainer container, final TaskMap tasks,
      final Map<Object, Object> context) throws ADLException {
    final Component[] subComponents = container.getComponents();

    final TaskMap.TaskHole[] subComponentTypes = new TaskMap.TaskHole[subComponents.length];
    final TaskMap.TaskHole[] subComponentInstances = new TaskMap.TaskHole[subComponents.length];
    for (int i = 0; i < subComponents.length; i++) {
      final Component subComponent = subComponents[i];
      subComponentTypes[i] = tasks.getTaskHole("type", subComponent);
      subComponentInstances[i] = tasks.getTaskHole("instance", subComponent);
    }
    createTask(container, tasks, subComponentTypes, subComponentInstances);
  }

  protected void createTask(final ComponentContainer container,
      final TaskMap tasks, final TaskMap.TaskHole[] subComponentTypes,
      final TaskMap.TaskHole[] subComponentInstances) {
    final AddContentTask task = new AddContentTask(builderItf,
        subComponentTypes, subComponentInstances);
    final TaskMap.TaskHole taskHole = addTask("content", task, container, tasks);
    getTaskHole("component", container, tasks)
        .addDependency(taskHole,
            RequireSourceCodeProviderTask.SOURCE_CODE_PROVIDER_TASK_ROLE,
            "content");
  }

  // --------------------------------------------------------------------------
  // Task class
  // --------------------------------------------------------------------------

  protected static class AddContentTask
      extends
        AbstractRecursiveSourceCodeProviderTask {
    private final ContentBuilder     builder;
    private final TaskMap.TaskHole[] subComponentTypes;
    private final TaskMap.TaskHole[] subComponentInstances;

    /**
     * @param builder
     * @param subComponentTypes
     * @param subComponentInstances
     */
    public AddContentTask(final ContentBuilder builder,
        final TaskMap.TaskHole[] subComponentTypes,
        final TaskMap.TaskHole[] subComponentInstances) {
      this.builder = builder;
      this.subComponentTypes = subComponentTypes;
      this.subComponentInstances = subComponentInstances;
      for (int i = 0; i < subComponentTypes.length; i++) {
        addPreviousTask(subComponentTypes[i]);
        addPreviousTask(subComponentInstances[i]);
      }
    }

    public void execute(final Map<Object, Object> context) throws Exception {
      final Object factory = getFactoryProviderTask().getFactory();
      final Object instance = (getInstanceProviderTask() == null)
          ? null
          : getInstanceProviderTask().getInstance();

      final Object[] types = new Object[subComponentTypes.length];
      final Object[] instances = new Object[subComponentTypes.length];
      for (int i = 0; i < subComponentTypes.length; i++) {
        types[i] = ((FactoryProviderTask) subComponentTypes[i].getTask())
            .getFactory();
        instances[i] = ((InstanceProviderTask) subComponentInstances[i]
            .getTask()).getInstance();
      }

      setResult(builder.addContent(factory, instance, types, instances,
          getCodePieces(), context));
    }
  }

  public void bindFc(final String s, final Object o)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException {

    if (s == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (s.equals(BUILDER_ITF)) {
      this.builderItf = (ContentBuilder) o;
    } else {
      throw new NoSuchInterfaceException("There is no interface named '" + s
          + "'");
    }
  }

  public String[] listFc() {
    return new String[]{BUILDER_ITF};
  }

  public Object lookupFc(final String s) throws NoSuchInterfaceException {

    if (s == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (s.equals(BUILDER_ITF)) {
      return this.builderItf;
    } else {
      throw new NoSuchInterfaceException("There is no interface named '" + s
          + "'");
    }
  }

  public void unbindFc(final String s) throws NoSuchInterfaceException,
      IllegalBindingException, IllegalLifeCycleException {
    if (s == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (s.equals(BUILDER_ITF)) {
      this.builderItf = null;
    } else {
      throw new NoSuchInterfaceException("There is no interface named '" + s
          + "'");
    }
  }

}
