/***
 * Cecilia ADL Compiler
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author:Ali Erdem Ozcan
 * Contributor: Alessio Pace
 */

package org.objectweb.fractal.cecilia.adl.interfaces;

import java.util.Map;

import org.objectweb.fractal.adl.components.ComponentContainer;
import org.objectweb.fractal.adl.implementations.Implementation;
import org.objectweb.fractal.adl.implementations.ImplementationContainer;
import org.objectweb.fractal.adl.types.TypeInterface;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.cecilia.adl.components.AbstractPrimitiveCompiler;
import org.objectweb.fractal.task.codegen.lib.AbstractRecursiveSourceCodeProviderTask;
import org.objectweb.fractal.task.core.TaskMap;

/**
 * Adds a task for the definition of interfaces.
 */
public abstract class AbstractInterfaceCompiler
    extends
      AbstractPrimitiveCompiler implements BindingController {

  // --------------------------------------------------------------------------
  // Client interface.
  // --------------------------------------------------------------------------

  /** the name of the client builder interface. */
  public static final String BUILDER_ITF = "builder";

  /**
   * The {@link InterfaceBuilder} used by this compiler to build definition of
   * every interfaces.
   */
  public InterfaceBuilder    builderItf;

  // --------------------------------------------------------------------------
  // Utility method
  // --------------------------------------------------------------------------

  protected TaskMap.TaskHole createTask(final String taskName,
      final TypeInterface[] interfaces, final ComponentContainer container,
      final TaskMap tasks) {

    String contentName = null;
    if (container instanceof ImplementationContainer) {
      final Implementation impl = ((ImplementationContainer) container)
          .getImplementation();
      if (impl != null) contentName = impl.getClassName();
    }

    final InterfaceTask task = new InterfaceTask(builderItf, contentName,
        interfaces);
    return addTask(taskName, task, container, tasks);
  }

  // --------------------------------------------------------------------------
  // Task class
  // --------------------------------------------------------------------------

  protected static class InterfaceTask
      extends
        AbstractRecursiveSourceCodeProviderTask {
    private final InterfaceBuilder builder;
    private final String           contentType;
    private final TypeInterface[]  interfaces;

    /**
     * @param builder
     * @param contentType
     * @param interfaces
     */
    public InterfaceTask(final InterfaceBuilder builder,
        final String contentType, final TypeInterface[] interfaces) {
      this.builder = builder;
      this.contentType = contentType;
      this.interfaces = interfaces;
    }

    public void execute(final Map<Object, Object> context) throws Exception {
      final Object factory = (getFactoryProviderTask() == null)
          ? null
          : getFactoryProviderTask().getFactory();
      final Object instance = (getInstanceProviderTask() == null)
          ? null
          : getInstanceProviderTask().getInstance();

      setResult(builder.createInterfaces(factory, instance, contentType,
          interfaces, getCodePieces(), context));
    }
  }

  // --------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // --------------------------------------------------------------------------

  public void bindFc(final String s, final Object o)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException {

    if (s == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (s.equals(BUILDER_ITF)) {
      this.builderItf = (InterfaceBuilder) o;
    } else {
      throw new NoSuchInterfaceException("There is no interface named '" + s
          + "'");
    }
  }

  public String[] listFc() {
    return new String[]{BUILDER_ITF};
  }

  public Object lookupFc(final String s) throws NoSuchInterfaceException {
    if (s == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (s.equals(BUILDER_ITF)) {
      return this.builderItf;
    } else {
      throw new NoSuchInterfaceException("There is no interface named '" + s
          + "'");
    }
  }

  public void unbindFc(final String s) throws NoSuchInterfaceException,
      IllegalBindingException, IllegalLifeCycleException {

    if (s == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (s.equals(BUILDER_ITF)) {
      this.builderItf = null;
    } else {
      throw new NoSuchInterfaceException("There is no interface named '" + s
          + "'");
    }
  }
}
