/***
 * Cecilia ADL Compiler
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.primitive.thinkMC.source;

import static org.objectweb.fractal.cecilia.adl.implementations.ImplementationDecorationUtil.getCode;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.CompilerError;
import org.objectweb.fractal.adl.Node;
import org.objectweb.fractal.adl.components.ComponentContainer;
import org.objectweb.fractal.adl.implementations.Implementation;
import org.objectweb.fractal.adl.implementations.ImplementationContainer;
import org.objectweb.fractal.adl.interfaces.Interface;
import org.objectweb.fractal.adl.interfaces.InterfaceContainer;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.cecilia.adl.directives.Include;
import org.objectweb.fractal.cecilia.adl.directives.IncludeContainer;
import org.objectweb.fractal.cecilia.adl.idl.ast.InterfaceDefinition;
import org.objectweb.fractal.cecilia.adl.idl.ast.InterfaceDefinitionContainer;
import org.objectweb.fractal.cecilia.adl.implementations.ImplementationDecorationUtil;
import org.objectweb.fractal.cecilia.adl.source.AbstractSourceFileCompiler;
import org.objectweb.fractal.cecilia.adl.source.SourceFileBuilder;
import org.objectweb.fractal.cecilia.adl.source.SourceFileProvider;
import org.objectweb.fractal.cecilia.adl.types.TypeDecorationUtil;
import org.objectweb.fractal.task.codegen.api.SourceFileProviderTask;
import org.objectweb.fractal.task.codegen.lib.SourceCodeAggregatorTask;
import org.objectweb.fractal.task.codegen.util.SourceFile;
import org.objectweb.fractal.task.core.Task;
import org.objectweb.fractal.task.core.TaskMap;
import org.objectweb.fractal.task.core.TaskMap.TaskHole;

/**
 * This compiler adds task for the generation of a source file containing the
 * definition of a component and all its instances. More over, it implements the
 * {@link SourceFileProvider} interface used by the <code>OutputCompiler</code>
 * to find the set of {@link SourceFileProviderTask} providing the source files
 * to be compiled.
 */
public class SourceFileCompiler extends AbstractSourceFileCompiler
    implements
      SourceFileProvider {

  // --------------------------------------------------------------------------
  // Client interface.
  // --------------------------------------------------------------------------

  /**
   * The name of the {@link #definitionSourceFileBuilderItf} client interface.
   */
  public static final String DEFINITION_BUILDER_ITF = "definition-builder";

  /**
   * The name of the {@link #moduleSourceFileBuilderItf} client interface.
   */
  public static final String MODULE_BUILDER_ITF     = "module-builder";

  /**
   * The name of the {@link #instancesSourceFileBuilderItf} client interface.
   */
  public static final String INSTANCE_BUILDER_ITF   = "instances-builder";

  /**
   * The {@link SourceFileBuilder} used by this compiler for multi-module
   * component. This builder generates the C-header file that defines the
   * component structure.
   */
  public SourceFileBuilder   definitionSourceFileBuilderItf;

  /**
   * The {@link SourceFileBuilder} used by this compiler for multi-module
   * component. This builder generates the C file that include one
   * implementation module.
   */
  public SourceFileBuilder   moduleSourceFileBuilderItf;

  /**
   * The {@link SourceFileBuilder} used by this compiler for multi-module
   * component. This builder generates the C file that contains the declaration
   * of the static instances.
   */
  public SourceFileBuilder   instancesSourceFileBuilderItf;

  // --------------------------------------------------------------------------
  // Implementation of the PrimitiveCompiler interface
  // --------------------------------------------------------------------------

  public void compile(final List<ComponentContainer> path,
      final ComponentContainer container, final TaskMap tasks,
      final Map<Object, Object> context) throws ADLException {

    // Search the InstancesAggregator task
    InstancesAggregator instancesTask;
    try {
      instancesTask = (InstancesAggregator) tasks.getTask("instances",
          TypeDecorationUtil.getTypeDecoration(container));
    } catch (final NoSuchElementException e) {
      if (TypeDecorationUtil.getTypeDecoration(container) != container) {
        throw new CompilerError("The component should defines a component type");
      }

      /*
       * the InstancesAggregator task does not exist, this means that the
       * source-task has not been created yet. Checks if the component is a
       * mono-module or a multi-module.
       */
      final List<Include> modules = getModules(container);
      if (modules == null) {
        // Mono module component
        instancesTask = createTasks(container, tasks);
      } else {
        // multi module component
        instancesTask = createTasks(container, modules, tasks);
      }

    }

    // Add the component instance code provider.
    instancesTask.addInstance(tasks
        .getTaskHole("component.instance", container));
  }

  // --------------------------------------------------------------------------
  // Utility methods
  // --------------------------------------------------------------------------

  /*
   * Returns the list of implementation module of the given container, or null
   * if the container has no implementation module (either a mono-module
   * primitive component or a composite component).
   */
  protected List<Include> getModules(final ComponentContainer container) {
    List<Include> modules = null;
    if (container instanceof ImplementationContainer) {
      final Implementation impl = ((ImplementationContainer) container)
          .getImplementation();
      if (impl != null) {
        if (impl instanceof IncludeContainer) {
          for (final Include include : ((IncludeContainer) impl).getIncludes()) {
            if (modules == null) modules = new ArrayList<Include>();
            modules.add(include);
          }
        }
      }
    }
    return modules;
  }

  /**
   * Create "source-file", "instances-source-file", and "module" tasks and
   * instances aggregator for multi-module tasks.
   */
  protected InstancesAggregator createTasks(final ComponentContainer container,
      final List<Include> modules, final TaskMap tasks) {
    final TaskHole typeTask = tasks.getTaskHole("type", container);

    // create the definition-source-file Task
    final SourceFileTask definitionTask = new SourceFileTask(
        definitionSourceFileBuilderItf);
    final TaskMap.TaskHole definitionTaskHole = tasks.addTask(
        "definition-source-file", container, definitionTask);
    definitionTask.setFactoryProviderTask(typeTask);
    definitionTask.addSourceCodeProviderTask("definition", tasks.getTaskHole(
        "component.definition", container));

    // Add dependency to data definition task
    final TaskMap.TaskHole dataDefinitionTask = tasks.getTaskHole(
        "implementation.definition", container);
    definitionTask.addSourceCodeProviderTask("data-definition",
        dataDefinitionTask);

    // create the InstancesAggregator task
    final InstancesAggregator instancesTask = new InstancesAggregator();
    final TaskMap.TaskHole instancesTaskHole = tasks.addTask("instances",
        container, instancesTask);

    // create the instance source file task
    final SourceFileTask instancesSourceFileTask = new SourceFileTask(
        instancesSourceFileBuilderItf);
    tasks.addTask("source-file", container, instancesSourceFileTask);

    // Add optional dependency to factory task
    final TaskMap.TaskHole factoryTaskOpt = tasks.getOptionalTaskHole(
        "component.factory", container);
    instancesSourceFileTask
        .addSourceCodeProviderTask("factory", factoryTaskOpt);

    instancesSourceFileTask.setFactoryProviderTask(typeTask);
    instancesSourceFileTask.addSourceCodeProviderTask("instances",
        instancesTaskHole);
    // add definitionTask as a previous task of instancesSourceFileTask to
    // ensure that the definitionTask is executed before the
    // instancesSourceFileTask is compiled.
    instancesSourceFileTask.addDependency(definitionTaskHole,
        Task.PREVIOUS_TASK_ROLE, null);

    // create module source file tasks
    for (final Include include : modules) {
      // retrieve source file associated to this include
      final SourceFile code = (SourceFile) getCode((Node) include);

      SourceFileProviderTask moduleSourceFileTask;
      if (code.isAssemblyFile()) {
        // if source file is an assembly file, moduleSourceFileTask is a
        // BasicSourceFileProviderTask.
        moduleSourceFileTask = new BasicSourceFileProviderTask(code);
      } else {
        // else moduleSourceFileTask must create a module source file.
        final SourceFileTask t = new SourceFileTask(moduleSourceFileBuilderItf);
        t.setFactoryProviderTask(typeTask);
        // add the "module" task as SourceCodeProvider
        t.addSourceCodeProviderTask("module", tasks.getTaskHole("module",
            include));
        // add definitionTask as a previous task of moduleSourceFileTask to
        // ensure that the definitionTask is executed before the
        // moduleSourceFileTask is compiled.
        t.addDependency(definitionTaskHole, Task.PREVIOUS_TASK_ROLE, null);
        moduleSourceFileTask = t;
      }
      tasks.addTask("module-source-file", include, moduleSourceFileTask);
    }

    return instancesTask;
  }

  /**
   * Create the "source-file" and the instances aggregator for mono-module
   * components.
   */
  protected InstancesAggregator createTasks(final ComponentContainer container,
      final TaskMap tasks) {
    final SourceFileTask sourceTask = new SourceFileTask(builderItf);
    tasks.addTask("source-file", container, sourceTask);

    sourceTask.setFactoryProviderTask(tasks.getTaskHole("type", container));
    sourceTask.addSourceCodeProviderTask("definition", tasks.getTaskHole(
        "component.definition", container));

    // Add optional dependency to implementation task
    final TaskMap.TaskHole implementationTaskOpt = tasks.getOptionalTaskHole(
        "implementation.definition", container);
    sourceTask.addSourceCodeProviderTask("implementation",
        implementationTaskOpt);

    // Add optional dependency to factory task
    final TaskMap.TaskHole factoryTaskOpt = tasks.getOptionalTaskHole(
        "component.factory", container);
    sourceTask.addSourceCodeProviderTask("factory", factoryTaskOpt);

    final InstancesAggregator instancesTask = new InstancesAggregator();
    final TaskMap.TaskHole instancesTaskHole = tasks.addTask("instances",
        container, instancesTask);
    // add the instancesTask has a source code provider of the source-file
    // task
    sourceTask.addSourceCodeProviderTask("instances", instancesTaskHole);
    return instancesTask;
  }

  // --------------------------------------------------------------------------
  // Implementation of the SourceFileProvider interface
  // --------------------------------------------------------------------------

  public Collection<TaskHole> getSourceFileProviderTask(
      final ComponentContainer container, final TaskMap tasks,
      final Map<Object, Object> context) {

    final List<TaskHole> sourceFileProviderTasks = new ArrayList<TaskHole>();
    final ComponentContainer typeContainer = TypeDecorationUtil
        .getTypeDecoration(container);
    sourceFileProviderTasks
        .add(tasks.getTaskHole("source-file", typeContainer));

    // find and add BasicSourceFileProviderTask for controller interface
    // implementation
    if (typeContainer instanceof InterfaceContainer) {
      for (final Interface itf : ((InterfaceContainer) typeContainer)
          .getInterfaces()) {
        final SourceFile src = (SourceFile) ImplementationDecorationUtil
            .getCode((Node) itf);
        if (src == null) continue;
        final TaskMap.TaskHole taskHole = tasks.getTaskHole("source-file", src);
        if (!taskHole.isFilled()) {
          final BasicSourceFileProviderTask task = new BasicSourceFileProviderTask(
              src);
          tasks.addTask("source-file", src, task);
        }
        if (itf instanceof InterfaceDefinitionContainer) {
          // add the IDL.definition task as previous task
          final InterfaceDefinition itfDef = ((InterfaceDefinitionContainer) itf)
              .getInterfaceDefinition();
          final TaskMap.TaskHole idlTaskHole = tasks.getTaskHole(
              "IDL.definition", itfDef.getName());
          taskHole.addDependency(idlTaskHole, Task.PREVIOUS_TASK_ROLE, null);
        }
        sourceFileProviderTasks.add(taskHole);
      }
    }
    final List<Include> modules = getModules(typeContainer);
    if (modules != null) {
      for (final Include module : modules) {
        sourceFileProviderTasks.add(tasks.getTaskHole("module-source-file",
            module));
      }
    }
    return sourceFileProviderTasks;
  }

  // --------------------------------------------------------------------------
  // Task classes
  // --------------------------------------------------------------------------

  protected static final class InstancesAggregator
      extends
        SourceCodeAggregatorTask {
    private final Set<TaskMap.TaskHole> instances = new HashSet<TaskMap.TaskHole>();

    /**
     * Default constructor.
     */
    public InstancesAggregator() {
      super("Component instances");
    }

    protected void addInstance(final TaskMap.TaskHole instanceCodeProviderTask) {
      // add the given task-hole only it has not been already added.
      if (instances.add(instanceCodeProviderTask)) {
        addSourceCodeProviderTask("instance" + instances.size(),
            instanceCodeProviderTask);
      }
    }
  }

  // --------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // --------------------------------------------------------------------------

  @Override
  public String[] listFc() {
    final String[] superList = super.listFc();
    final String[] list = new String[superList.length + 3];
    System.arraycopy(superList, 0, list, 0, superList.length);
    list[superList.length] = DEFINITION_BUILDER_ITF;
    list[superList.length + 1] = MODULE_BUILDER_ITF;
    list[superList.length + 2] = INSTANCE_BUILDER_ITF;
    return list;
  }

  @Override
  public void bindFc(final String s, final Object o)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException {

    if (s == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (s.equals(DEFINITION_BUILDER_ITF)) {
      this.definitionSourceFileBuilderItf = (SourceFileBuilder) o;
    } else if (s.equals(MODULE_BUILDER_ITF)) {
      this.moduleSourceFileBuilderItf = (SourceFileBuilder) o;
    } else if (s.equals(INSTANCE_BUILDER_ITF)) {
      this.instancesSourceFileBuilderItf = (SourceFileBuilder) o;
    } else {
      super.bindFc(s, o);
    }
  }

  @Override
  public Object lookupFc(final String s) throws NoSuchInterfaceException {
    if (s == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (s.equals(DEFINITION_BUILDER_ITF)) {
      return this.definitionSourceFileBuilderItf;
    } else if (s.equals(MODULE_BUILDER_ITF)) {
      return this.moduleSourceFileBuilderItf;
    } else if (s.equals(INSTANCE_BUILDER_ITF)) {
      return this.instancesSourceFileBuilderItf;
    } else {
      return super.lookupFc(s);
    }
  }

  @Override
  public void unbindFc(final String s) throws NoSuchInterfaceException,
      IllegalBindingException, IllegalLifeCycleException {

    if (s == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (s.equals(DEFINITION_BUILDER_ITF)) {
      this.definitionSourceFileBuilderItf = null;
    } else if (s.equals(MODULE_BUILDER_ITF)) {
      this.moduleSourceFileBuilderItf = null;
    } else if (s.equals(INSTANCE_BUILDER_ITF)) {
      this.instancesSourceFileBuilderItf = null;
    } else {
      super.unbindFc(s);
    }
  }
}
